import requests
import inspect

from mumuki.i18n import Locale

class BaseMumuki():
    def __init__(self, token:str, locale:str, url = "https://mumuki.io"):
        self._token = token
        self._url = url
        self._locale = Locale(locale)
        self._solution = None
        self._connected = False
        self._register_globals()

    def visit(self, organization:str, exercise):
        self._prepare_before_visit()

        self.__organization = organization
        self._exercise_id = int(exercise)

        if self._offline():
            return

        try:
            self._connect_after_visit()
            self._connected = True
        except:
            self._report_auth_error()

    def _solution_url(self):
        return f"{self._exercise_url()}/solutions"

    def _exercise_url(self):
        return f"{self._url}/{self.__organization}/exercises/{self._exercise_id}"

    def _headers(self):
        return { "Authorization": f"Bearer {self._token}" }

    def _get_exercise(self):
        return requests.get(
             self._exercise_url(),
             headers = self._headers())

    def _post_solution(self):
        return requests.post(
            self._solution_url(),
            json = { "solution": { "content": self._solution } },
            headers = self._headers())

    def register_solution(self, function=None):
        self._solution = self._get_source(function)

    def get_solution(self):
        return self._solution

    def test(self, function=None):
        self._prepare_before_test()
        if self._solution is None:
            self.register_solution(function)
        self._submit()

    def _submit(self):
        if self._offline():
            return

        if not self._connected:
            self._report_auth_error()
            return

        result = self._post_solution()
        if result.status_code == 403:
            self._report_auth_error()
        else:
            self._report_results(result.json())

    def _get_source(self, function=None):
        if type(function) == str:
            return function
        if function is not None:
            lines = inspect.getsourcelines(function)[0][1:]
            indent_size = inspect.indentsize("".join(line.rstrip("\n\t ") for line in lines))
            return "".join(line[indent_size:] for line in lines)
        else:
            return self._source_missing()

    def _report_auth_error(self):
        print(self._locale.translate("could_not_load_exercise", id = self._exercise_id, url = self._exercise_url()))



import pytest

from mumuki import Mumuki

def test_can_create_with_default_url():
    mumuki = Mumuki("token", "es")
    assert mumuki._url == "https://mumuki.io"

def test_can_create_with_custom_url():
    mumuki = Mumuki("token", "es", "http://localhost:3000")
    assert mumuki._url == "http://localhost:3000"

def test_not_offline_by_default():
    mumuki = Mumuki("token", "es")
    assert not mumuki._offline()

def test_can_register_plain_string():
    mumuki = Mumuki("token", "es")
    mumuki.register_solution("def foo(): pass")
    assert mumuki._solution == "def foo(): pass"

def test_can_register_function():
    def solution():
        def baz(): return 4 + 5

    mumuki = Mumuki("token", "es")
    mumuki.register_solution(solution)
    assert mumuki._solution == "def baz(): return 4 + 5\n"



locales = {
  "en": {
    "could_not_load_exercise": "Could not access exercise {id}. Please visit {url} and verify the instructions",
    "magic_cell_missing": "Please ensure to mark you solution cell with %%solution",
    "no_ipython": "There is not ipython environment available. Make sure you are running from a Jupyter or Colab notebook"
  },
  "es": {
    "could_not_load_exercise": "No se pudo acceder al ejercicio {id}. Visitá {url} y verificá las instrucciones",
    "magic_cell_missing": "Asegurate de marcar tu celda de solución con %%solution",
    "no_ipython": "No hay un entorno de ipython disponible. Asegurate de estar en un cuaderno Jupyter o Colab"
  },
  "pt": {
    "could_not_load_exercise": "Não foi possível acessar o exercício {id}. Visite {url} e verifique as instruções",
    "magic_cell_missing": "Certifique-se de marcar sua célula de solução com %%solution",
    "no_ipython": "Não há ambiente ipython disponível. Verifique se você está executando em um caderno Jupyter ou Colab"
  }
}


class Locale:
  def __init__(self, code: str):
    if code not in locales:
      raise RuntimeError(f"Invalid locale {code}")

    self.code = code

  def translate(self, key: str, **kwargs) -> str:
    return locales[self.code][key].format(**kwargs)



import sys
import inspect

from mumuki.base import BaseMumuki

class Mumuki(BaseMumuki):
    def _prepare_before_visit(self):
        frame = inspect.stack()[2]
        self.__start_line = frame.lineno
        self.__file = frame.filename

    def _register_globals(self):
        pass

    def _source_missing(self):
        with open(self.__file) as f:
            return "".join(f.readlines()[self.__start_line:self.__end_line-1])

    def _connect_after_visit(self):
        self._get_exercise()
        print("OK")

    def _prepare_before_test(self):
        frame = inspect.stack()[2]
        self.__end_line = frame.lineno

    def _report_results(self, results):
        print(results['status'])

    def _offline(self):
        return sys.flags.interactive


import pytest

from mumuki.interactive import IMumuki, soup
from bs4 import BeautifulSoup

from IPython.testing.globalipapp import get_ipython
import pytest

def setup_module():
    get_ipython()

def test_can_create_with_default_url():
    mumuki = IMumuki("token", "es")
    assert mumuki._url == "https://mumuki.io"

def test_reports_failed_visits_en(capsys):
    mumuki = IMumuki("token", "en", "http://localhost")
    mumuki.visit("nonexistent", 100)

    captured = capsys.readouterr()
    assert captured.out == "Could not access exercise 100. Please visit http://localhost/nonexistent/exercises/100 and verify the instructions\n"

def test_reports_failed_visits_es(capsys):
    mumuki = IMumuki("token", "es", "http://localhost")
    mumuki.visit("nonexistent", 100)

    captured = capsys.readouterr()
    assert captured.out == "No se pudo acceder al ejercicio 100. Visitá http://localhost/nonexistent/exercises/100 y verificá las instrucciones\n"

def test_reports_failed_visits_pt(capsys):
    mumuki = IMumuki("token", "pt", "http://localhost")
    mumuki.visit("nonexistent", 100)

    captured = capsys.readouterr()
    assert captured.out == "Não foi possível acessar o exercício 100. Visite http://localhost/nonexistent/exercises/100 e verifique as instruções\n"



def test_is_never_offline():
    mumuki = IMumuki("token", "es")
    assert not mumuki._offline()

def test_test_fails_if_there_is_no_magic_cell_en():
    mumuki = IMumuki("token", "en")
    with pytest.raises(RuntimeError) as e:
      mumuki.test()

    assert "Please ensure to mark you solution cell with %%solution" in str(e.value)

def test_test_fails_if_there_is_no_magic_cell_es():
    mumuki = IMumuki("token", "es")
    with pytest.raises(RuntimeError) as e:
      mumuki.test()

    assert "Asegurate de marcar tu celda de solución con %%solution" in str(e.value)


def test_test_fails_if_there_is_no_magic_cell_pt():
    mumuki = IMumuki("token", "pt")
    with pytest.raises(RuntimeError) as e:
      mumuki.test()

    assert "Certifique-se de marcar sua célula de solução com %%solution" in str(e.value)


def test_makes_links_open_in_new_window():
    body = soup(
        """
        <a class="btn btn-complementary w-100" role="button"
          href="https://mumuki.io/central/exercises/1"><span
            class="fa5-text-r">Siguiente Ejercicio: Ejemplo</span><i
            class="fas fa-chevron-right"></i>
        </a>
        """
    )
    assert body == BeautifulSoup(
        """
        <a class="btn btn-complementary w-100" role="button"
          href="https://mumuki.io/central/exercises/1" target="_blank"><span
            class="fa5-text-r">Siguiente Ejercicio: Ejemplo</span><i
            class="fas fa-chevron-right"></i>
        </a>
        """,
        "html.parser",
    )


def test_leaves_download_links_as_is():
    body = soup(
        """
        <a download="solution.py"
          href="data:application/octet-stream,content">
          <i class="fas fa-download"></i>
          <span class="fa5-text">Descargá lo que hiciste</span>
        </a>
        """
    )
    assert body == BeautifulSoup(
        """
        <a download="solution.py"
          href="data:application/octet-stream,content">
          <i class="fas fa-download"></i>
          <span class="fa5-text">Descargá lo que hiciste</span>
        </a>
        """,
        "html.parser",
    )



from bs4 import BeautifulSoup
from IPython.core.display import HTML
from IPython.display import display
from IPython.core.getipython import get_ipython

from mumuki.base import BaseMumuki

def soup(html):
    s = BeautifulSoup(html, "html.parser")
    for link in s.find_all("a"):
        if 'download' not in link.attrs:
            link['target'] = '_blank'
    return s

def html(soup: BeautifulSoup):
    return HTML(str(soup))

class IMumuki(BaseMumuki):
    def show(self):
        s = soup(self._get_exercise().text)
        display(html(s.body.find_all("h1")[0]))
        display(html(s.body.find_all("div", {"class":"exercise-assignment"})[0]))

    def _register_globals(self):
        def solution(_line, cell):
            self.register_solution(cell)
            get_ipython().run_cell(cell)

        ip = get_ipython()
        if ip:
            ip.register_magic_function(solution, 'cell', 'solution')
        else:
            raise RuntimeError(self._locale.translate("no_ipython"))

    def _source_missing(self):
        raise RuntimeError(self._locale.translate("magic_cell_missing"))

    def _prepare_before_visit(self):
        pass

    def _connect_after_visit(self):
        self.show()

    def _prepare_before_test(self):
        pass

    def _report_results(self, results):
        display(html(soup(results["html"])))

    def _offline(self):
        return False
